@@ -22,6 +22,7 @@ from .util import default_openw, default_mkdirs, sep_from_open, ParquetException
 
 MARKER = b'PAR1'
 NaT = np.timedelta64(None).tobytes()  # require numpy version >= 1.7
+nat = np.datetime64('NaT').view('int64')
 
 typemap = {  # primitive type, converted type, bit width
     'bool': (parquet_thrift.Type.BOOLEAN, None, 1),
@@ -42,7 +43,7 @@ revmap = {parquet_thrift.Type.INT32: np.int32,
           parquet_thrift.Type.FLOAT: np.float32}
 
 
-def find_type(data, convert=False):
+def find_type(data, convert=False, fixed_text=None):
     """ Get appropriate typecodes for column dtype
 
     Data conversion may happen here, only at write time.
@@ -51,11 +52,8 @@ def find_type(data, convert=False):
     before saving to parquet, we will not make any assumptions for them.
 
     If the dtype is "object" the first ten items will be examined, and is str
-    or bytes, will be stored as variable length byte strings; if dict or list,
-    (nested data) will be stored with JSON encoding.
-
-    To be stored as fixed-length byte strings, the dtype must be "bytesXX"
-    (pandas notation) or "|SXX" (numpy notation)
+    or bytes, will be stored as variable or fixed length byte strings;
+    if dict or list, (nested data) will be stored with JSON encoding.
 
     In the case of catagoricals, the data type refers to the labels; the data
     (codes) will be stored as int. The labels are usually variable length
@@ -69,12 +67,18 @@ def find_type(data, convert=False):
 
     Parameters
     ----------
-    A pandas series.
+    data: pd.Series
+    convert: bool (False)
+        Whether to actually perform the conversion, or just infer types
+    fixed_text: int or None
+        For str and bytes, the fixed-string length to use. If None, object
+        column will remain variable length.
 
     Returns
     -------
     - a thrift schema element
     - a thrift typecode to be passed to the column chunk writer
+    - converted data (None if convert is False)
 
     """
     out = None
@@ -82,7 +86,7 @@ def find_type(data, convert=False):
     if dtype.name in typemap:
         type, converted_type, width = typemap[dtype.name]
         if type in revmap and convert:
-            out = data.values.astype(revmap[type])
+            out = data.values.astype(revmap[type], copy=False)
         elif type == parquet_thrift.Type.BOOLEAN and convert:
             padded = np.lib.pad(data.values, (0, 8 - (len(data) % 8)),
                                 'constant', constant_values=(0, 0))
@@ -98,14 +102,29 @@ def find_type(data, convert=False):
     elif dtype == "O":
         head = data[:10] if isinstance(data, pd.Index) else data.valid()[:10]
         if all(isinstance(i, str) for i in head):
-            type, converted_type, width = (parquet_thrift.Type.BYTE_ARRAY,
-                                           parquet_thrift.ConvertedType.UTF8, None)
-            if convert:
-                out = data.str.encode('utf8').values
+            if fixed_text:
+                width = fixed_text
+                if convert:
+                    out = data.str.encode('utf8').values.astype('S%i' % width)
+                type, converted_type = (
+                    parquet_thrift.Type.FIXED_LEN_BYTE_ARRAY,
+                    parquet_thrift.ConvertedType.UTF8)
+            else:
+                type, converted_type, width = (parquet_thrift.Type.BYTE_ARRAY,
+                                               parquet_thrift.ConvertedType.UTF8, None)
+                if convert:
+                    out = data.str.encode('utf8').values
         elif all(isinstance(i, bytes) for i in head):
-            type, converted_type, width = parquet_thrift.Type.BYTE_ARRAY, None, None
-            if convert:
-                out = data.values
+            if fixed_text:
+                width = fixed_text
+                if convert:
+                    out = data.values.astype('S%i' % width)
+                type, converted_type = (
+                    parquet_thrift.Type.FIXED_LEN_BYTE_ARRAY, None)
+            else:
+                type, converted_type, width = parquet_thrift.Type.BYTE_ARRAY, None, None
+                if convert:
+                    out = data.values
         elif all(isinstance(i, (list, dict)) for i in head):
             type, converted_type, width = (parquet_thrift.Type.BYTE_ARRAY,
                                            parquet_thrift.ConvertedType.JSON, None)
@@ -119,12 +138,14 @@ def find_type(data, convert=False):
         if hasattr(dtype, 'tz') and str(dtype.tz) != 'UTC':
             warnings.warn('Coercing datetimes to UTC')
         if convert:
-            out = data.values.astype('datetime64[us]')
+            out = np.empty(len(data), 'int64')
+            time_shift(data.values.view('int64'), out)
     elif str(dtype).startswith("timedelta64"):
         type, converted_type, width = (parquet_thrift.Type.INT64,
                                        parquet_thrift.ConvertedType.TIME_MICROS, None)
         if convert:
-            out = data.values.astype('timedelta64[us]')
+            out = np.empty(len(data), 'int64')
+            time_shift(data.values.view('int64'), out)
     else:
         raise ValueError("Don't know how to convert data type: %s" % dtype)
     # TODO: pandas has no explicit support for Decimal
@@ -134,6 +155,15 @@ def find_type(data, convert=False):
     return se, type, out
 
 
+@numba.njit()
+def time_shift(indata, outdata, factor=1000):
+    for i in range(len(indata)):
+        if indata[i] == nat:
+            outdata[i] = nat
+        else:
+            outdata[i] = indata[i] // factor
+
+
 def thrift_print(structure, offset=0):
     """
     Handy recursive text ouput for thrift structures
@@ -189,10 +219,10 @@ def write_thrift(fobj, thrift):
     return fobj.tell() - t0
 
 
-def encode_plain(data, se):
+def encode_plain(data, se, fixed_text=None):
     """PLAIN encoding; returns byte representation"""
-    se, type, out = find_type(data, True)
-    if data.dtype == "O":
+    se, type, out = find_type(data, True, fixed_text=fixed_text)
+    if se.type == parquet_thrift.Type.BYTE_ARRAY:
         return b''.join([struct.pack('<l', len(x)) + x for x in out])
     else:
         return out.tobytes()
@@ -287,12 +317,13 @@ def encode_rle_bp(data, width, o, withlength=False):
         o.loc = end
 
 
-def encode_dict(data, se):
+def encode_dict(data, se, _):
     """ The data part of dictionary encoding is always int32s, with RLE/bitpack
     """
+    # TODO: should width be a parameter equal to len(cats) ?
     width = encoding.width_from_max_int(data.max())
     ldata = ((len(data) + 7) // 8) * width + 11
-    i = data.values.astype(np.int32)
+    i = data.values.astype(np.int32, copy=False)
     out = encoding.Numpy8(np.empty(ldata, dtype=np.uint8))
     out.write_byte(width)
     encode_rle_bp(i, width, out)
@@ -306,28 +337,29 @@ encode = {
 }
 
 
-def make_definitions(data):
+def make_definitions(data, no_nulls):
     """For data that can contain NULLs, produce definition levels binary
     data: either bitpacked bools, or (if number of nulls == 0), single RLE
     block."""
-    valid = ~data.isnull()
     temp = encoding.Numpy8(np.empty(10, dtype=np.uint8))
 
-    if valid.all():
+    if no_nulls:
         # no nulls at all
-        l = len(valid)
+        l = len(data)
         encode_unsigned_varint(l << 1, temp)
         temp.write_byte(1)
         block = struct.pack('<i', temp.loc) + temp.so_far().tostring()
+        out = data
     else:
         # bitpack bools
-        out = encode_plain(valid, None)
+        out = encode_plain(data.notnull(), None)
 
         encode_unsigned_varint(len(out) << 1 | 1, temp)
         head = temp.so_far().tostring()
 
         block = struct.pack('<i', len(head + out)) + head + out
-    return block, data.valid()
+        out = data.valid()  # better, data[data.notnull()], from above ?
+    return block, out
 
 
 def write_column(f, data, selement, encoding='PLAIN', compression=None):
@@ -352,19 +384,19 @@ def write_column(f, data, selement, encoding='PLAIN', compression=None):
 
     """
     has_nulls = selement.repetition_type == parquet_thrift.FieldRepetitionType.OPTIONAL
+    fixed_text = selement.type_length
     tot_rows = len(data)
 
-    # no NULL handling (but NaNs, NaTs are allowed)
     if has_nulls:
-        print('has nulls!')
-        definition_data, data = make_definitions(data)
+        num_nulls = data.count() - len(data)
+        definition_data, data = make_definitions(data, num_nulls == 0)
     else:
         definition_data = b""
+        num_nulls = 0
 
     # No nested field handling (encode those as J/BSON)
     repetition_data = b""
 
-    rows = len(data)
     cats = False
     name = data.name
     diff = 0
@@ -373,7 +405,8 @@ def write_column(f, data, selement, encoding='PLAIN', compression=None):
         dph = parquet_thrift.DictionaryPageHeader(
                 num_values=len(data.cat.categories),
                 encoding=parquet_thrift.Encoding.PLAIN)
-        bdata = encode['PLAIN'](pd.Series(data.cat.categories), selement)
+        bdata = encode['PLAIN'](pd.Series(data.cat.categories), selement,
+                                fixed_text=fixed_text)
         l0 = len(bdata)
         if compression:
             bdata = compress_data(bdata, compression)
@@ -389,20 +422,28 @@ def write_column(f, data, selement, encoding='PLAIN', compression=None):
         dict_start = f.tell()
         write_thrift(f, ph)
         f.write(bdata)
+        try:
+            max, min = data.max(), data.min()
+            max = encode['PLAIN'](pd.Series([max]), selement,
+                                  fixed_text=fixed_text)
+            min = encode['PLAIN'](pd.Series([min]), selement,
+                                  fixed_text=fixed_text)
+        except TypeError:
+            max, min = None, None
         data = data.cat.codes.astype(np.int32)
         cats = True
         encoding = "PLAIN_DICTIONARY"
 
     start = f.tell()
-    bdata = definition_data + repetition_data + encode[encoding](data, selement)
+    bdata = definition_data + repetition_data + encode[encoding](data, selement,
+                                                                 fixed_text)
     try:
-        max, min = data.max(), data.min()
-        if encoding == "DELTA_BINARY_PACKED":
-            encode2 = "PLAIN"
-        else:
-            encode2 = encoding
-        max = encode[encode2](pd.Series([max], dtype=data.dtype), selement)
-        min = encode[encode2](pd.Series([min], dtype=data.dtype), selement)
+        if encoding != 'PLAIN_DICTIONARY':
+            max, min = data.max(), data.min()
+            max = encode['PLAIN'](pd.Series([max], dtype=data.dtype), selement,
+                                  fixed_text=fixed_text)
+            min = encode['PLAIN'](pd.Series([min], dtype=data.dtype), selement,
+                                  fixed_text=fixed_text)
     except TypeError:
         max, min = None, None
 
@@ -432,7 +473,7 @@ def write_column(f, data, selement, encoding='PLAIN', compression=None):
     uncompressed_size = compressed_size + diff
 
     offset = f.tell()
-    s = parquet_thrift.Statistics(max=max, min=min, null_count=0)
+    s = parquet_thrift.Statistics(max=max, min=min, null_count=num_nulls)
 
     p = [parquet_thrift.PageEncodingStats(
             page_type=parquet_thrift.PageType.DATA_PAGE,
@@ -484,7 +525,8 @@ def make_row_group(f, data, schema, file_path=None, compression=None,
     return rg
 
 
-def make_part_file(f, data, schema, compression=None, encoding='PLAIN'):
+def make_part_file(f, data, schema, compression=None, encoding='PLAIN',
+                   fixed_text=None):
     if len(data) == 0:
         return
     with f as f:
@@ -502,7 +544,7 @@ def make_part_file(f, data, schema, compression=None, encoding='PLAIN'):
     return rg
 
 
-def make_metadata(data, has_nulls=[], ignore_columns=[]):
+def make_metadata(data, has_nulls=True, ignore_columns=[], fixed_text=None):
     root = parquet_thrift.SchemaElement(name='schema',
                                         num_children=0)
 
@@ -515,12 +557,19 @@ def make_metadata(data, has_nulls=[], ignore_columns=[]):
     for column in data.columns:
         if column in ignore_columns:
             continue
+        fixed = None if fixed_text is None else fixed_text.get(column, None)
         if str(data[column].dtype) == 'category':
-            se, type, _ = find_type(data[column].cat.categories)
+            se, type, _ = find_type(data[column].cat.categories,
+                                    fixed_text=fixed)
             se.name = column
         else:
-            se, type, _ = find_type(data[column])
-        if column in has_nulls and str(data[column].dtype) in ['category', 'object']:
+            se, type, _ = find_type(data[column], fixed_text=fixed)
+        if has_nulls is None:
+            se.repetition_type = type == parquet_thrift.Type.BYTE_ARRAY
+        else:
+            has_nulls = (has_nulls if has_nulls in [True, False]
+                         else column in has_nulls)
+        if has_nulls and data[column].dtype.kind != 'i':
             se.repetition_type = parquet_thrift.FieldRepetitionType.OPTIONAL
         fmd.schema.append(se)
         root.num_children += 1
@@ -529,8 +578,8 @@ def make_metadata(data, has_nulls=[], ignore_columns=[]):
 
 def write(filename, data, row_group_offsets=50000000, encoding="PLAIN",
           compression=None, file_scheme='simple', open_with=default_openw,
-          mkdirs=default_mkdirs, has_nulls=[], write_index=None,
-          partition_on=[]):
+          mkdirs=default_mkdirs, has_nulls=None, write_index=None,
+          partition_on=[], fixed_text=None):
     """ Write Pandas DataFrame to filename as Parquet Format
 
     Parameters
@@ -559,10 +608,14 @@ def write(filename, data, row_group_offsets=50000000, encoding="PLAIN",
         When called with a path/URL, creates any necessary dictionaries to
         make that location writable, e.g., ``os.makedirs``. This is not
         necessary if using the simple file scheme
-    has_nulls: list of strings
-        The named columns can have nulls. Only applies to Object and Category
-        columns, as pandas ints can't have NULLs, and NaN/NaT is equivalent
-        to NULL in float and time-like columns.
+    has_nulls: None, bool or list of strings
+        Whether columns can have nulls. If a list of strings, those given
+        columns will be marked as "optional" in the metadata, and include
+        null definition blocks on disk. Some data types (floats and times)
+        can instead use the sentinel values NaN and NaT, which are not the same
+        as NULL in parquet, but functionally act the same in many cases,
+        particularly if converting back to pandas later. A value of None
+        will assume nulls for object columns and not otherwise.
     write_index: boolean
         Whether or not to write the index to a separate column.  By default we
         write the index *if* it is not 0, 1, ..., n.
@@ -570,6 +623,11 @@ def write(filename, data, row_group_offsets=50000000, encoding="PLAIN",
         Passed to groupby in order to split data within each row-group,
         producing a structured directory tree. Note: as with pandas, null
         values will be dropped. Ignored if file_scheme is simple.
+    fixed_text: {column: int length} or None
+        For bytes or str columns, values will be converted
+        to fixed-length strings of the given length for the given columns
+        before writing, potentially providing a large speed
+        boost.
 
     Examples
     --------
@@ -596,7 +654,8 @@ def write(filename, data, row_group_offsets=50000000, encoding="PLAIN",
                 data.index._step == 1 and data.index.name is None):
             data = data.reset_index()
         ignore = partition_on if file_scheme != 'simple' else []
-        fmd = make_metadata(data, has_nulls=has_nulls, ignore_columns=ignore)
+        fmd = make_metadata(data, has_nulls=has_nulls, ignore_columns=ignore,
+                            fixed_text=fixed_text)
         for i, start in enumerate(row_group_offsets):
             end = row_group_offsets[i+1] if i < (len(row_group_offsets) - 1) else None
             if file_scheme == 'simple':
